"""
composer.py
"""
from django import template
from django.conf import settings
from django.core.urlresolvers import reverse

from ..models import Element


register = template.Library()


class LazyContent:
    """
    Wrapper around a `NodeList` that is lazily-evaluated to a string.
    """
    def __init__(self, nodelist, context):
        self.nodelist = nodelist
        self.context = context

    def __str__(self):
        return self.nodelist.render(self.context)


def serialize_context(context):
    """
    Serializes a dictionary, list, or object into an easy to read example of the
    context.
    """
    if type(context) == dict:
        return '\n'.join(['%s = %r' % (key, val) for key, val in context.items()])

    if type(context) == list:
        result = 'list of %d elements' % len(context)
        for i, e in enumerate(context[0:15]):
            result += '\n%2d = %r' % (i, e)
        if len(context) > 15:
            result += '\n ...\n%2d = %r' % (len(context), context[-1])
        return result

    if context is not None or context != {}:
        result = 'object: ' + repr(context)

    return ''


class ComposerNode(template.Node):

    def __init__(self, name, content=None, nodelist=None, context_var=None, dynamic=False):
        self.name = name
        self.content = content
        self.nodelist = nodelist
        self.dynamic = dynamic
        if context_var is not None:
            self.context_var = template.Variable(context_var)
        else:
            self.context_var = None

    def render(self, context):
        """
        The `template_name` context variable must be present.
        """
        if self.nodelist:
            self.content = LazyContent(self.nodelist, context)

        template = context.get('template_name', None)
        elements = context.get('composer_elements', None)
        if template is None or elements is None:
            return str(self.content)

        has_perm = context.get('can_compose_permission', False)

        el_context = {}
        if self.context_var is not None:
            el_context = self.context_var.resolve(context)

        if self.name not in elements:
            context['composer_elements'][self.name] = Element.objects.create(
                template_name=template,
                name=self.name,
                is_dynamic=self.dynamic,
                context_example=serialize_context(el_context),
                content=str(self.content))

        element = elements[self.name]
        try:
            result = element.render(self.content, el_context)
        except Exception as e:
            if has_perm:
                result = str(e)
            else:
                result = element.content

        if has_perm:
            url = reverse('composer-edit-element', kwargs={'pk': element.id})
            result = '<div class="edit-composer-button">edit</div>' + result
            result = '<div class="edit-composer-element" data-url="%s" data-name="%s">%s</div>' % (
                url, self.name, result)

        return result


def strip_quotes(val):
    quotes = ("'", '"')
    if type(val) == str and val[0] in quotes and val[-1] in quotes:
        return val[1:-1]
    return val


@register.tag(name='composer_dynamic')
def do_composer_dynamic(parser, token):
    """
    This tag expects the following format:

        {% composer_dynamic 'name' context_var %}
            This is the default template, which can use {% if blah %}context {{vars}}{% endif %}
        {% endcomposer %}

    The second argument is optional.
    """
    bits = token.split_contents()
    name, context_var = None, None
    try:
        if len(bits) == 2:
            _, name = bits
        else:
            _, name, context_var = bits
    except ValueError:
        raise template.TemplateSyntaxError(
            '%r tag requires either 1 or 2 arguments' % token.contents.split()[0])

    name = strip_quotes(name)
    nodelist = parser.parse(('endcomposer',))
    parser.delete_first_token()
    return ComposerNode(name, nodelist=nodelist, context_var=context_var, dynamic=True)


@register.tag(name='composer_static')
def do_composer_static(parser, token):
    """
    This tag expects the following format:

        {% composer_static 'name' %}Some static default content.{% endcomposer %}
    """
    try:
        _, name = token.split_contents()
    except ValueError:
        raise template.TemplateSyntaxError(
            '%r tag requires exactly 1 argument' % token.contents.split()[0])

    name = strip_quotes(name)
    nodelist = parser.parse(('endcomposer',))
    parser.delete_first_token()
    return ComposerNode(name, nodelist=nodelist)


@register.simple_tag(takes_context=True)
def composer_includes(context):
    """
    Include the composer JS and CSS files in a page if the user has permission.
    """
    if context.get('can_compose_permission', False):
        url = settings.STATIC_URL
        url += '' if url[-1] == '/' else '/'
        js = '<script type="text/javascript" src="%sjs/composer.min.js"></script>' % url
        css = '<link rel="stylesheet" type="text/css" href="%scss/composer.css">' % url
        return js + css
    return ''



"""
models.py
"""
import datetime

from django.db import models
from django.conf import settings
from django.template import Template, Context, TemplateSyntaxError


class Element(models.Model):
    """
    Contains text/template to be injected into pages.
    """
    template_name = models.CharField(max_length=200)
    name = models.CharField(max_length=100)

    is_dynamic = models.BooleanField(default=False)
    last_changed = models.DateTimeField(auto_now=True)
    has_changed = models.BooleanField(default=False)
    changed_by = models.ForeignKey(settings.AUTH_USER_MODEL, null=True, blank=True)
    context_example = models.TextField(null=True, blank=True)
    content = models.TextField(null=True, blank=True)

    class Meta:
        app_label = 'composer'
        index_together = ('name', 'template_name')
        unique_together = ('name', 'template_name')

    class TemplateError(Exception):
        pass

    def save(self, *args, **kwargs):
        """
        Update timestamp automatically.
        """
        self.last_changed = datetime.datetime.now()
        return super().save(*args, **kwargs)

    def render(self, default_content, context=None):
        """
        Dynamic templates are actually rendered using the Django template
        engine, while static templates are just returned as-is.

        The default_content will be displayed as-is if this element hasn't been
        changed before.
        """
        if not self.has_changed:
            return str(default_content)
        
        if not self.is_dynamic:
            return self.content

        tpl = Template(self.content)
        return tpl.render(Context(context or {}))

    def attempt_update(self, new_content, changed_by):
        """
        Attempt to update this element with new content.
        """
        self.has_changed = True

        if self.is_dynamic:
            try:
                Template(new_content).render(Context({}))
            except TemplateSyntaxError as e:
                raise self.TemplateError(str(e))

        self.content = new_content
        self.changed_by = changed_by
        self.has_changed = True
        self.save()



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
from django.conf import settings


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Element',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('template_name', models.CharField(max_length=200)),
                ('name', models.CharField(max_length=100)),
                ('is_dynamic', models.BooleanField(default=False)),
                ('last_changed', models.DateTimeField(auto_now=True)),
                ('has_changed', models.BooleanField(default=False)),
                ('context_example', models.TextField(null=True, blank=True)),
                ('content', models.TextField(null=True, blank=True)),
                ('changed_by', models.ForeignKey(blank=True, to=settings.AUTH_USER_MODEL, null=True)),
            ],
        ),
        migrations.AlterUniqueTogether(
            name='element',
            unique_together=set([('name', 'template_name')]),
        ),
        migrations.AlterIndexTogether(
            name='element',
            index_together=set([('name', 'template_name')]),
        ),
    ]



"""
urls.py
"""
from django.conf.urls import patterns, url
from django.conf import settings

from .views import EditComposerElementView, ExampleView


urlpatterns = patterns(
    '',

    url(r'edit/(?P<pk>[\d]+)$', EditComposerElementView.as_view(), name='composer-edit-element')
)

if settings.DEBUG:
    urlpatterns += (
        url(r'example/', ExampleView.as_view(), name='composer-example'),
    )



"""
views.py
"""
from django.views.generic import TemplateView
from django.views.generic.detail import SingleObjectMixin
from django.conf import settings
from django.core.exceptions import PermissionDenied

from .models import Element


def default_can_compose_permission(view, request, template_name):
    """
    Default global permission check.
    """
    return request.user and request.user.is_authenticated and (
        request.user.is_staff or request.user.is_superuser)


class ComposerPermissionMixin:
    def can_compose_permission(self, request, template_name):
        """
        Can be overridden at the view level to provide view specific permission
        checks. Otherwise, the global permission check is permformed.
        """
        perm = getattr(settings, 'COMPOSER_CAN_COMPOSE_PERMISSION', default_can_compose_permission)
        return perm(self, request, template_name)


class ComposerViewMixin(ComposerPermissionMixin):
    """
    View mixin for any Django CBV that uses `django.views.generic.base.ContextMixin`
    and `django.views.views.generic.TemplateResponseMixin`.
    """

    def get_context_data(self, **context):
        """
        Get the available compose elements for this view if allowed.
        """
        context = super().get_context_data(**context)

        template = self.get_template_names()[0]

        context['can_compose_permission'] = self.can_compose_permission(self.request, template)
        context['template_name'] = template
        context['composer_elements'] = {e.name: e
                                        for e in Element.objects.filter(template_name=template)}

        return context


class EditComposerElementView(ComposerPermissionMixin, SingleObjectMixin, TemplateView):
    """
    Ajax friendly view for editing a composer element.
    """
    model = Element
    template_name = 'composer/edit.html'

    def dispatch(self, request, *args, **kwargs):
        """
        Only allow users with permission.
        """
        self.object = self.get_object()
        if not self.can_compose_permission(request, self.object.template_name):
            raise PermissionDenied

        return super().dispatch(request, *args, **kwargs)

    def get_context_data(self, **context):
        """
        Add the composer element.
        """
        context['element'] = self.object

        if self.object.is_dynamic and self.object.context_example is not None:
            context['context_example'] = '<br>'.join(self.object.context_example.split('\n'))

        if 'content' not in context:
            context['content'] = self.object.content

        return context

    def post(self, request, *args, **kwargs):
        """
        Update the contents if they render correctly.
        """
        ctx = {}
        status = 200
        content = request.POST.get('content', '')
        
        try:
            self.object.attempt_update(content, request.user)
        except self.model.TemplateError as e:
            ctx = {'error': str(e), 'content': content}
            status = 400

        return self.render_to_response(self.get_context_data(**ctx), status=status)


class ExampleView(ComposerViewMixin, TemplateView):
    """
    Shows an example of different usages of composer elements on a static page.
    """
    template_name = 'composer/example.html'



import os

from setuptools import setup

README = open(os.path.join(os.path.dirname(__file__), 'README.rst')).read()
VERSION = '0.0.4'

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))


def get_packages(package):
    """
    Return root package and all sub-packages.
    """
    return [dirpath
            for dirpath, dirnames, filenames in os.walk(package)
            if os.path.exists(os.path.join(dirpath, '__init__.py'))]


def get_package_data(package):
    """
    Return all files under the root package, that are not in a
    package themselves.
    """
    walk = [(dirpath.replace(package + os.sep, '', 1), filenames)
            for dirpath, dirnames, filenames in os.walk(package)
            if not os.path.exists(os.path.join(dirpath, '__init__.py'))]

    filepaths = []
    for base, filenames in walk:
        filepaths.extend([os.path.join(base, filename)
                          for filename in filenames])
    return {package: filepaths}


setup(
    name='django-composer',

    version=VERSION,

    packages=get_packages('composer'),
    package_data=get_package_data('composer'),

    license='MIT License',

    description='Simple Django app for adding dynamic text and markup to templates and views.',
    long_description=README,

    keywords='',

    url='https://github.com/featherweightweb/django-composer',
    download_url='https://github.com/featherweightweb/django-composer/tarball/%s' % VERSION,

    author='Benjamin Dummer',
    author_email='ben@featherweightweb.com',

    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4'
    ],
)
